
#ifndef PLS_INVOKE_PARALLEL_IMPL_H
#define PLS_INVOKE_PARALLEL_IMPL_H

#include "pls/internal/scheduling/fork_join_task.h"
#include "pls/internal/scheduling/scheduler.h"

namespace pls {
    namespace algorithm {
        namespace internal {
            using namespace ::pls::internal::scheduling;

            template<typename Body>
            inline void run_body(const Body& internal_body, const abstract_task::id& id) {
                // Make sure we are in the context of this invoke_parallel instance,
                // if not we will spawn it as a new 'fork-join-style' task.
                auto current_task = scheduler::current_task();
                if (current_task->unique_id() == id) {
                    auto current_sub_task = reinterpret_cast<fork_join_task*>(current_task)->currently_executing();
                    internal_body(current_sub_task);
                } else {
                    fork_join_lambda<Body> root_body(&internal_body);
                    fork_join_task root_task{&root_body, id};
                    scheduler::execute_task(root_task);
                }
            }
        }

        template<typename Function1, typename Function2>
        void invoke_parallel(const Function1& function1, const Function2& function2) {
            using namespace ::pls::internal::scheduling;
            static abstract_task::id id{PLS_UNIQUE_ID, true};

            auto internal_body = [&] (fork_join_sub_task* this_task){
                auto sub_task_body_1 = [&] (fork_join_sub_task*){ function1(); };
                auto sub_task_1 = fork_join_lambda<decltype(sub_task_body_1)>(&sub_task_body_1);

                this_task->spawn_child(sub_task_1);
                function2(); // Execute last function 'inline' without spawning a sub_task object
                this_task->wait_for_all();
            };

            internal::run_body(internal_body, id);
        }

        template<typename Function1, typename Function2, typename Function3>
        void invoke_parallel(const Function1& function1, const Function2& function2, const Function3& function3) {
            using namespace ::pls::internal::scheduling;
            static abstract_task::id id{PLS_UNIQUE_ID, true};

            auto internal_body = [&] (fork_join_sub_task* this_task){
                auto sub_task_body_1 = [&] (fork_join_sub_task*){ function1(); };
                auto sub_task_1 = fork_join_lambda<decltype(sub_task_body_1)>(&sub_task_body_1);
                auto sub_task_body_2 = [&] (fork_join_sub_task*){ function2(); };
                auto sub_task_2 = fork_join_lambda<decltype(sub_task_body_2)>(&sub_task_body_2);

                this_task->spawn_child(sub_task_1);
                this_task->spawn_child(sub_task_2);
                function3(); // Execute last function 'inline' without spawning a sub_task object
                this_task->wait_for_all();
            };

            internal::run_body(internal_body, id);
        }
    }
}

#endif //PLS_INVOKE_PARALLEL_IMPL_H
